/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * The announcement message queue is a PUB/SUB queue the daemon binds to and
 * the clients connect to.  The daemon then publishes a message every time
 * the QSO index changes.  In other words, it sends a message every time a
 * QSO is added, removed, or modified.
 *
 * The notification message is very simple.  It consists of a 32-bit big
 * endian message type (see enum announce_msg) followed by the 16-byte uuid
 * identifying the QSO generating the notification.
 */

#include <zmq.h>

#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/cstr.h>

#include "announce.h"

#define SOCK_NAME	"sock.announce"

static void *sock;
static void *ctx;

int announce_start(void)
{
	char endpoint[PATH_MAX];
	int ret;

	snprintf(endpoint, sizeof(endpoint), "ipc://%s", SOCK_NAME);

	ctx = zmq_ctx_new();
	if (!ctx) {
		cmn_err(CE_ERROR, "failed to create zmq context: %s",
			strerror(errno));
		return -errno;
	}

	sock = zmq_socket(ctx, ZMQ_PUB);
	if (!sock) {
		cmn_err(CE_ERROR, "failed to create zmq socket: %s",
			strerror(errno));
		ret = -errno;
		goto err;
	}

	if (zmq_bind(sock, endpoint)) {
		cmn_err(CE_ERROR, "failed to bind zmq socket: %s",
			strerror(errno));
		ret = -errno;
		goto err_sock;
	}

	cmn_err(CE_INFO, "Bound to: ipc://%s (MQ)", SOCK_NAME);

	return 0;

err_sock:
	if (zmq_close(sock))
		cmn_err(CE_ERROR, "failed to close zmq socket: %s",
			strerror(errno));

err:
	if (zmq_ctx_destroy(ctx))
		cmn_err(CE_ERROR, "failed to destroy zmq context: %s",
			strerror(errno));

	return ret;
}

void announce_stop(void)
{
	if (zmq_close(sock))
		cmn_err(CE_ERROR, "failed to close zmq socket: %s",
			strerror(errno));

	if (zmq_ctx_destroy(ctx))
		cmn_err(CE_ERROR, "failed to destroy zmq context: %s",
			strerror(errno));
}

static void publish(enum announce_msg_type type, struct announce_msg *msg)
{
	int ret;

	msg->type = cpu32_to_be(type);
	msg->_pad0 = 0;

	ret = zmq_send(sock, msg, sizeof(*msg), 0);
	if (ret < 0)
		cmn_err(CE_WARN, "announce: failed to send notification: %s",
			strerror(errno));
	else if (ret != sizeof(*msg))
		cmn_err(CE_WARN, "announce: failed to send the entire "
			"notification of %zu bytes (%d sent)",
			sizeof(*msg), ret);
}

void announce_notify_qso(enum announce_msg_type type, const struct xuuid *uuid)
{
	struct announce_msg msg;

	memcpy(msg.u.qso.uuid, uuid->raw, sizeof(uuid->raw));

	publish(type, &msg);
}

void announce_notify_loc(enum announce_msg_type type, uint8_t mode, char *grid)
{
	struct announce_msg msg;

	/* make sure mode and grid match and that mode is valid */
	switch (mode) {
		case 0:
			if (grid && strlen(grid))
				goto err_inval_grid;
			grid = "";
			break;
		case 2:
		case 3:
			if (!grid || !strlen(grid) || (strlen(grid) % 2 != 0))
				goto err_inval_grid;
			break;
		default:
			cmn_err(CE_WARN, "Ignoring location notification "
				"with invalid mode=%u", mode);
			return;
	}

	/* make sure the message type and mode match */
	switch (type) {
		case ANNOUNCE_LOC_ACQ:
			if ((mode != 2) && (mode != 3))
				goto err_inval_mode;
			break;
		case ANNOUNCE_LOC_CHANGE:
		case ANNOUNCE_LOC_CURRENT:
			/* any of the modes is ok */
			break;
		case ANNOUNCE_LOC_LOST:
		case ANNOUNCE_LOC_DISCONNECT:
			if (mode != 0)
				goto err_inval_mode;
			break;
		case ANNOUNCE_QSO_NEW:
		case ANNOUNCE_QSO_UPDATED:
			panic("invalid type specified (%d)", type);
	}

	msg.u.loc.mode = mode;
	strcpy_safe((char *) msg.u.loc.grid, grid, sizeof(msg.u.loc.grid));
	memset(msg.u.loc._pad1, 0, sizeof(msg.u.loc._pad1));

	publish(type, &msg);
	return;

err_inval_grid:
	cmn_err(CE_WARN, "Ignoring location notification with invalid grid "
		"(msg %d, mode %d, grid %p, len %zu)", type, mode, grid,
		grid ? strlen(grid) : 0);
	return;

err_inval_mode:
	cmn_err(CE_WARN, "Ignoring location notification with invalid mode "
		"(msg %d, mode %d)", type, mode);
}
