/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/version.h>
#include <jeffpc/error.h>
#include <jeffpc/io.h>
#include <jeffpc/thread.h>

#include <locale.h>

#include <hlog/country.h>
#include <hlog/index.h>
#include <hlog/version.h>

#include "announce.h"
#include "location.h"
#include "wsjtx.h"

#define DATA_DIR_COUNTRY_FNAME	"cty.csv"

static int load_countries(void)
{
	int ret;

	cmn_err(CE_INFO, "Loading countries data...");
	ret = country_load(AT_FDCWD, DATA_DIR_COUNTRY_FNAME);
	cmn_err(CE_INFO, "%s.", ret ? "error" : "done");

	if (ret)
		cmn_err(CE_ERROR, "Error: failed to load country data: %s",
			xstrerror(ret));

	return ret;
}

static int load_index(void)
{
	int ret;

	cmn_err(CE_INFO, "Loading existing QSOs...");
	ret = index_init(AT_FDCWD);
	cmn_err(CE_INFO, "%s.", ret ? "error" : "done");

	if (ret)
		cmn_err(CE_ERROR, "Error: failed to load existing entries: %s",
			xstrerror(ret));

	return ret;
}

static void usage(const char *prog)
{
	cmn_err(CE_ERROR, "Usage: %s <path to data dir>", prog);
	exit(1);
}

int main(int argc, char **argv)
{
	const char *data_dir;
	int ret;

	setlocale(LC_ALL, "");
	ASSERT0(putenv("TZ=UTC"));

	xthr_set_name("main");

	cmn_err(CE_INFO, "hlogd version %s", version_string);
	cmn_err(CE_INFO, "libjeffpc version %s", jeffpc_version);

	if (argc != 2)
		usage(argv[0]);

	data_dir = argv[1];

	/*
	 * chdir to the data directory to make our lives in the future much
	 * easier.  We will be able to use relative paths everywhere.
	 */
	if (chdir(data_dir)) {
		cmn_err(CE_ERROR, "Error: failed to change to data dir: %s",
			strerror(errno));
		goto err;
	}

	ret = load_countries();
	if (ret)
		goto err;

	ret = load_index();
	if (ret)
		goto err;

	ret = announce_start();
	if (ret)
		goto err;

	ret = location_start();
	if (ret)
		goto err_announce;

	ret = wsjtx_start();
	if (ret)
		goto err_location;

	for (;;)
		sleep(60);

	location_stop();
	announce_stop();

	return 0;

err_location:
	location_stop();

err_announce:
	announce_stop();

err:
	return 1;
}
