/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <hlog/qso.h>
#include <hlog/util.h>
#include "xlua_impl.h"
#include "wrapper.h"

/*
 * Expose a struct qso to lua scripts
 */

/* no actual ref counting */
static inline void qso_side_putref(struct qso_side *side)
{
}

struct field {
	const char *name;
	enum {
		T_SIDE,
		T_QSO_VAL,
		T_QSO_TIME,
		T_QSO_UUID,
		T_QSO_SIDE,
		T_QSO_ADDL,
	} type;

	union {
		struct {
			struct val *(*get)(const struct qso_side *);
			int (*set)(struct qso_side *, const char *);
		} side; /* T_SIDE */
		struct {
			struct val *(*get)(const struct qso *);
			int (*set)(struct qso *, const char *);
		} qso_val; /* T_QSO_VAL */
		struct {
			struct time *(*get)(struct qso *);
		} qso_time; /* T_QSO_TIME */
		struct {
			bool tx;
		} qso_side; /* T_QSO_SIDE */
	};
};

static struct time *__get_start_time(struct qso *qso);
static struct time *__get_end_time(struct qso *qso);
static struct val *__get_power(const struct qso_side *side);

struct field side_fields[] = {
	{ "station_call", T_SIDE,
		.side = { qso_get_station_call_val, qso_set_station_call } },
	{ "operator_call", T_SIDE,
		.side = { qso_get_operator_call_val, qso_set_operator_call } },
	{ "mode", T_SIDE,
		.side = { qso_get_mode_val, qso_set_mode } },
	{ "power", T_SIDE,
		.side = { __get_power, qso_set_power_user } },
	{ "freq", T_SIDE,
		.side = { qso_get_freq, qso_set_freq } },
	{ "band", T_SIDE,
		.side = { qso_get_band, NULL } },
	{ "rst", T_SIDE,
		.side = { qso_get_rst_val, qso_set_rst } },
	{ "name", T_SIDE,
		.side = { qso_get_name_val, qso_set_name } },
	{ "qth", T_SIDE,
		.side = { qso_get_qth_val, qso_set_qth } },
	{ "grid", T_SIDE,
		.side = { qso_get_grid_val, qso_set_grid } },
	{ "country", T_SIDE,
		.side = { qso_get_country_val, qso_set_country } },
	{ "state", T_SIDE,
		.side = { qso_get_state_val, qso_set_state } },
	{ "county", T_SIDE,
		.side = { qso_get_county_val, qso_set_county } },
	{ "continent", T_SIDE,
		.side = { qso_get_continent_val, qso_set_continent } },
	{ "dxcc", T_SIDE,
		.side = { qso_get_zone_dxcc, qso_set_zone_dxcc_user } },
	{ "itu", T_SIDE,
		.side = { qso_get_zone_itu, qso_set_zone_itu_user } },
	{ "cqz", T_SIDE,
		.side = { qso_get_zone_cqz, qso_set_zone_cqz_user } },
	{ "seq", T_SIDE,
		.side = { qso_get_seq, qso_set_seq_user } },
};

struct field qso_fields[] = {
	{ "uuid", T_QSO_UUID },
	{ "start_time", T_QSO_TIME,
		.qso_time = { __get_start_time } },
	{ "end_time", T_QSO_TIME,
		.qso_time = { __get_end_time } },
	{ "comment", T_QSO_VAL,
		.qso_val = { qso_get_comment_val, qso_set_comment } },
	{ "tx", T_QSO_SIDE,
		.qso_side = { true } },
	{ "rx", T_QSO_SIDE,
		.qso_side = { false } },
	{ "additional", T_QSO_ADDL },
};

static int xlua_qso_index_data(lua_State *L);
static int xlua_qso_newindex_data(lua_State *L);

static const luaL_Reg __qso_fxns[] = {
	{ "__index", xlua_qso_index_data },
	{ "__newindex", xlua_qso_newindex_data },
	{ NULL, NULL },
};

static int xlua_qso_side_index_data(lua_State *L);
static int xlua_qso_side_newindex_data(lua_State *L);

static const luaL_Reg __qso_side_fxns[] = {
	{ "__index", xlua_qso_side_index_data },
	{ "__newindex", xlua_qso_side_newindex_data },
	{ NULL, NULL },
};

WRAP_C_DATA(qso, struct qso, qso_putref, __qso_fxns)
WRAP_C_DATA(qso_side, struct qso_side, qso_side_putref, __qso_side_fxns)

static struct time *__get_start_time(struct qso *qso)
{
	return &qso->start;
}

static struct time *__get_end_time(struct qso *qso)
{
	return &qso->end;
}

static struct val *__get_power(const struct qso_side *side)
{
	char str[30];

	if (!pretty_print_float(str, sizeof(str), side->power, 1000, 3))
		return NULL;

	return VAL_DUP_STR(str);
}

static void __push_val(lua_State *L, struct val *val)
{
	if (!val) {
		lua_pushnil(L);
		return;
	}

	switch (val->type) {
		case VT_STR:
			lua_pushstring(L, val_cstr(val));
			goto out;
		case VT_INT:
			lua_pushinteger(L, val->i);
			goto out;
		case VT_BOOL:
		case VT_BLOB:
		case VT_ARRAY:
		case VT_NVL:
		case VT_CHAR:
		case VT_NULL:
		case VT_SYM:
		case VT_CONS:
			panic("unsupported type %s", val_typename(val->type));
	}

	panic("unknown type %d (%s)", val->type, val_typename(val->type));

out:
	val_putref(val);
}

static int xlua_index_data(lua_State *L, const struct field *fields,
			   size_t nfields, void *data, const char *name,
			   bool ro, bool is_side)
{
	size_t i;

	/* push the return values */
	for (i = 0; i < nfields; i++) {
		const struct field *field = &fields[i];

		if (strcmp(name, field->name))
			continue;

		switch (field->type) {
			case T_SIDE:
				if (!is_side)
					panic("programming error");
				break;
			case T_QSO_VAL:
			case T_QSO_TIME:
			case T_QSO_UUID:
			case T_QSO_SIDE:
			case T_QSO_ADDL:
				if (is_side)
					panic("programming error");
				break;
		}

		switch (field->type) {
			case T_SIDE:
				__push_val(L, field->side.get(data));
				break;
			case T_QSO_VAL:
				__push_val(L, field->qso_val.get(data));
				break;
			case T_QSO_TIME:
				xlua_pushtime(L, field->qso_time.get(data),
					      true);

				/* set struct time's uservalue == struct qso */
				lua_pushvalue(L, 1 /* qso index */);
				lua_setuservalue(L, -2 /* time index */);
				break;
			case T_QSO_UUID: {
				struct qso *qso = data;

				lua_pushstring(L, qso->uuid_str);
				break;
			}
			case T_QSO_SIDE: {
				struct qso *qso = data;
				struct qso_side *side;

				if (field->qso_side.tx)
					side = &qso->tx;
				else
					side = &qso->rx;

				xlua_pushqso_side(L, side, ro);

				/* set qso side's uservalue == struct qso */
				lua_pushvalue(L, 1 /* qso index */);
				lua_setuservalue(L, -2 /* side index */);
				break;
			}
			case T_QSO_ADDL: {
				struct qso *qso = data;

				xlua_pushnvl(L, nvl_getref(qso->additional), ro);
				break;
			}
		}

		return 1;
	}

	return luaL_error(L, "unknown field name '%s'", name);
}

/* handle a qso[key], where table is -2 and key is -1 [-2,+1,e] */
static int xlua_qso_index_data(lua_State *L)
{
	bool ro;

	xlua_checkstack(L, 2, 2, 1);

	return xlua_index_data(L, qso_fields, ARRAY_LEN(qso_fields),
			       xlua_qso_getptr(L, -2, &ro),
			       luaL_checkstring(L, -1),
			       ro, false);
}

/* handle a qso_side[key], where table is -2 and key is -1 [-2,+1,e] */
static int xlua_qso_side_index_data(lua_State *L)
{
	bool ro;

	xlua_checkstack(L, 2, 2, 1);

	return xlua_index_data(L, side_fields, ARRAY_LEN(side_fields),
			       xlua_qso_side_getptr(L, -2, &ro),
			       luaL_checkstring(L, -1),
			       ro, true);
}

static int xlua_newindex_data(lua_State *L, const struct field *fields,
			      size_t nfields, void *data, const char *name,
			      bool ro, bool is_side)
{
	size_t i;

	/* prevent modifications to read-only structures */
	if (ro)
		return luaL_error(L, "cannot change read-only QSO");

	for (i = 0; i < nfields; i++) {
		const struct field *field = &fields[i];
		const char *val;
		int ret;

		if (strcmp(name, field->name))
			continue;

		switch (field->type) {
			case T_SIDE:
				if (!is_side)
					panic("programming error");
				break;
			case T_QSO_VAL:
			case T_QSO_TIME:
			case T_QSO_UUID:
			case T_QSO_SIDE:
			case T_QSO_ADDL:
				if (is_side)
					panic("programming error");
				break;
		}

		switch (field->type) {
			case T_SIDE:
				if (!field->side.set)
					goto readonly;

				val = luaL_checkstring(L, -1);

				ret = field->side.set(data, val);
				break;
			case T_QSO_VAL:
				if (!field->qso_val.set)
					goto readonly;

				val = luaL_checkstring(L, -1);

				ret = field->qso_val.set(data, val);
				break;
			case T_QSO_ADDL:
			case T_QSO_TIME:
			case T_QSO_UUID:
			case T_QSO_SIDE:
				goto readonly;
		}

		if (ret)
			luaL_error(L, "failed to set field: %s",
				   xstrerror(ret));

		return 0;
	}

	return luaL_error(L, "unknown field name '%s'", name);

readonly:
	return luaL_error(L, "cannot change read-only field '%s'", name);
}

/*
 * handle a qso[key] = value, where table is -3, key is -2, and value is -1
 * [-3,+0,e]
 */
static int xlua_qso_newindex_data(lua_State *L)
{
	bool ro;

	xlua_checkstack(L, 3, 3, 0);

	return xlua_newindex_data(L, qso_fields, ARRAY_LEN(qso_fields),
				  xlua_qso_getptr(L, -3, &ro),
				  luaL_checkstring(L, -2),
				  ro, false);
}

/*
 * handle a qso_side[key] = value, where table is -3, key is -2, and value is -1
 * [-3,+0,e]
 */
static int xlua_qso_side_newindex_data(lua_State *L)
{
	bool ro;

	xlua_checkstack(L, 3, 3, 0);

	return xlua_newindex_data(L, side_fields, ARRAY_LEN(side_fields),
				  xlua_qso_side_getptr(L, -3, &ro),
				  luaL_checkstring(L, -2),
				  ro, true);
}

/* get struct qso from the stack [-0,+0,e] */
struct qso *xlua_toqso(lua_State *L, int idx)
{
	return xlua_qso_getptr(L, idx, NULL);
}

/*
 * Call a function with the qso as an argument.  The passed in qso ref is
 * consumed.  If ro==true, struct qso cannot be modified by the lua script.
 *
 * Returns:
 *  -ENOENT: function not found
 */
int xlua_qso_call(lua_State *L, const char *funcname, struct qso *qso,
		  bool ro, char *error, size_t errlen)
{
	const int orig_top = lua_gettop(L) - (funcname ? 0 : 1);
	int ret;

	if (funcname)
		ret = lua_getglobal(L, funcname);
	else
		ret = lua_type(L, -1);

	if (ret != LUA_TFUNCTION) {
		lua_pop(L, 1);
		ASSERT_LUA_STACK(L, orig_top);
		qso_putref(qso);

		strcpy_safe(error, "not a function", errlen);

		return -ENOENT;
	}

	xlua_pushqso(L, qso, ro);

	/*
	 * now invoke the function
	 */
	ret = lua_pcall(L, 1, 0, 0);
	if (ret)
		COPY_ERROR(L, -1, error, errlen);

	ASSERT_LUA_STACK(L, orig_top);

	return lua2errno(ret);
}

/* allocate a new struct qso [-0,+1,e] */
static int xlua_qso_new(lua_State *L)
{
	struct qso *qso;

	xlua_checkstack(L, 0, 0, 1);

	qso = qso_alloc();
	if (!qso)
		return luaL_error(L, "failed to allocate struct qso");

	xlua_pushqso(L, qso, false);

	return 1;
}

static const luaL_Reg qso_funcs[] = {
	{ "new", xlua_qso_new },
	{ NULL, NULL },
};

void xlua_pushlib_qso(lua_State *L)
{
	luaL_newlib(L, qso_funcs);
}
